import java.util.List;

public class ColoredGraph<T> extends Graph<T> {

    public void greedyColoring() {
        // Einfärbungen zurücksetzen
        for (Node<T> node : getNodes()) {
            node.setColor(null);
        }

        for (Node<T> node : getNodes()) {
            boolean[] usedColors = new boolean[getNodes().size()];

            for (Node<T> neighbor : node.getNeighbors()) {
                Integer color = (Integer) neighbor.getColor();
                if (color != null) {
                    usedColors[color] = true;
                }
            }

            for (int color = 0; color < usedColors.length; color++) {
                if (!usedColors[color]) {
                    node.setColor(color);
                    break;
                }
            }
        }
    }

    public void welshPowellColoring() {
        // Einfärbungen zurücksetzen
        for (Node<T> node : getNodes()) {
            node.setColor(null);
        }

        // Knoten nach Gradanzahl absteigend sortieren
        List<Node<T>> sortedNodes = getNodes().stream()
                .sorted((n1, n2) -> Integer.compare(n2.getDegree(), n1.getDegree()))
                .toList();

        Integer color = 0; // erste Farbe, z.B. 0 = rot, 1 = blau, 2 = gelb, 3 = grün

        // Farbzuweisung beginnend mit Knoten mit höchsten Grad
        for (int i = 0; i < sortedNodes.size(); i++) {

            Node<T> node = sortedNodes.get(i);
            if (node.getColor() == null) { // Knoten noch nicht eingefärbt
                node.setColor(color);

                // Färbe nicht benachbarte Knoten mit derselben Farbe
                for (int j = i + 1; j < sortedNodes.size(); j++) {

                    Node<T> otherNode = sortedNodes.get(j);
                    if (otherNode.getColor() == null &&
                            !areNeighbors(node, otherNode) &&
                            !areSameColoredNeighbors(otherNode.getNeighbors(), color)) {
                        otherNode.setColor(color);
                    }
                }
                color++; // nächste Farbe
            }
        }
    }

    public boolean backtrackingColoring() {
        // Setze alle Farben zurück
        for (Node<T> node : getNodes()) {
            node.setColor(null);
        }

        int maxColors = 0; // Beginne mit 1 Farbe und erhöhe schrittweise
        while (!colorGraphBacktracking(0, maxColors)) {
            maxColors++; // Erhöhe die Anzahl der Farben, bis eine Lösung gefunden wird
        }
        return true;
    }

    private boolean colorGraphBacktracking(int nodeIndex, int maxColors) {
        if (nodeIndex == getNodes().size()) {
            return true; // Alle Knoten wurden erfolgreich gefärbt
        }

        Node<T> node = getNodes().get(nodeIndex);
        for (int color = 0; color <= maxColors; color++) {
            if (!areSameColoredNeighbors(node.getNeighbors(), color)) {
                node.setColor(color);

                if (colorGraphBacktracking(nodeIndex + 1, maxColors)) {
                    return true; // Rekursiver Aufruf erfolgreich
                }

                node.setColor(null); // Backtracking
            }
        }
        return false; // Keine gültige Färbung mit maxColor gefunden
    }

    private boolean areNeighbors(Node<T> node1, Node<T> node2) {
        return node1.getNeighbors().contains(node2) ||
                node2.getNeighbors().contains(node1);
    }

    private boolean areSameColoredNeighbors(List<Node<T>> neighbors, Integer color) {
        for (Node<T> neighbor : neighbors) {
            if (color == neighbor.getColor()) {
                return true;
            }
        }
        return false;
    }

    public String generateDotRepresentation() {
        // Graph aus DOT-Language erstellen: https://magjac.com/graphviz-visual-editor/
        // oder über GraphViw->Download (Kommandozeilenprogramm dot.exe)
        // Farben: https://graphviz.org/doc/info/colors.html
        String[] colors = { "red", "yellow", "green", "cyan", "pink", "aliceblue", "wheat", "violet", "snow", "plum1" };
        StringBuilder dotCode = new StringBuilder(getDirected() ? "digraph G {\n" : "graph G {\n");

        for (Node<T> node : getNodes()) {
            String node1 = node.getItem() != null ? node.getItem().toString() : "";

            if (node.getColor() != null) {
                dotCode.append(String.format("  \"%s\" [shape=circle, style=filled, fillcolor=%s];\n",
                        node1, colors[node.getColor()]));
            }

            for (int i = 0; i < node.getNeighbors().size(); i++) {
                Node<T> neighbor = node.getNeighbors().get(i);
                String node2 = neighbor.getItem() != null ? neighbor.getItem().toString() : "";
                Double weight = node.getWeights().get(i);

                String edge = String.format("  \"%s\" -> \"%s\" [label=\"%s\"];\n",
                        node1, node2, (weight != null ? weight.toString() : " "));

                if (!getDirected()) { // umgekehrte gleiche Kanten ausfiltern
                    String edgeInverted = String.format("  \"%s\" -> \"%s\" [label=\"%s\"];\n",
                            node2, node1, (weight != null ? weight.toString() : " "));
                    if (dotCode.toString().contains(edgeInverted)) {
                        continue;
                    }
                }
                dotCode.append(edge);
            }
        }

        if (!getDirected()) {
            return dotCode.toString().replace("->", "--") + "}";
        }

        return dotCode.append("}").toString();
    }
}